## Copyright (C) 2012 pdiribarne
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with Octave; see the file COPYING.  If not, see
## <http://www.gnu.org/licenses/>.

## -*- texinfo -*-
## @deftypefn  {Function File} {@var{h} =} struct2hdl (@var{s})
## @deftypefnx {Function File} {@var{h} =} struct2hdl (@var{s}, @var{p})
## Constructs a object from the structure @var{s}. The structure must
## contain the fields; "handle", "type", "children", "properties", and
## "special".  If the handle of an existing figure or axes is specified,
## @var{p}, the new object will be created as a child to that object.
## If no object handle is provided, then a new figure and the necessary
## children will be constructed using the default object values from
## the root figure.
## @seealso{findobj, get, hdl2struct, set}
## @end deftypefn  

## Author: pdiribarne <pdiribarne@new-host.home>
## Created: 2012-03-04

function [ h, matchout ] = struct2hdl (hgS, matchin=[])
  fields = { "handle", "type", "children", "properties", "special"};
  if nargin > 2 || ! isstruct (hgS)
    print_usage ();
  elseif !all (isfield (hgS, fields))
    print_usage ();
  elseif isempty (matchin)

  elseif isscalar (matchin)
    if !ishandle (matchin)
      error ("struct2hdl: argument #2 is not a handle to graphic object")
    endif
    if strcmp (get (matchin).type, "figure")
      if strcmp (hgS.type, "figure")
        error ("struct2hdl: incompatible inputs")
      endif
      cf = gcf;
      figure (matchin);
    elseif strcmp (get (matchin).type, "axes")
      if strcmp (hgS.type, "figure") || strcmp (hgS.type, "axes")
        error ("struct2hdl: incompatible inputs")
      endif
      cf = gcf;
      ax = gca;

      fig = get (matchin, "parent");
      figure (fig);
      set (fig, "currentaxes", matchin);
    else
      print_usage ();
    endif
    matchin = [NaN; matchin];
  endif

  ## main object
  if strcmpi (hgS.type, "figure")
    h = figure ();
  elseif strcmpi (hgS.type, "axes")
    [h, hgS] = createaxes (hgS, matchin);
  elseif strcmpi (hgS.type, "line")
    h = createline (hgS);
  elseif strcmpi (hgS.type, "patch")
    h = createpatch (hgS);
  elseif strcmpi (hgS.type, "text")
    h = createtext (hgS);
  elseif strcmpi (hgS.type, "image")
    h = createimage (hgS);
  elseif strcmpi (hgS.type, "surface")
    h = createsurface (hgS);
  elseif strcmpi (hgS.type, "hggroup")
    [h, hgS, matchin] = createhg (hgS, matchin);
  endif

  ## children
  if ! isempty (h)
    matchout = [matchin [hgS.handle; h]]; # correspondance [original; new] handle identifier
    ## regular children
    temp = matchout;
    ch = hgS.children;
    nch = length (ch);
    ii = 0;
    while nch
      ii++;
      if !any (ii == hgS.special)
        [h2, matchout] = struct2hdl (hgS.children(ii), matchout);
      endif
      nch--;
    endwhile

  else
    matchout = matchin;
  endif

  ## paste properties
  setprops (hgS, h, matchout);

endfunction ##struct2hdl

function [h, hgSout] = createaxes (hgS, match);

  ## regular axes
  if strcmpi (hgS.properties.tag, "")
    propval = {"position", hgS.properties.position};
    ##propval = {};
    hid = {"autopos_tag", "looseinset"};
    for ii = 1:numel (hid)
      prop = hid{ii};
      if isfield (hgS.properties, prop)
        val = hgS.properties.(prop);
        propval = [propval, prop, val];
      endif
    endfor
    h = axes (propval{:});
    hgS.properties = rmfield (hgS.properties, hid);

  ## legends
  elseif strcmpi (hgS.properties.tag, "legend")
    oldax = hgS.properties.userdata.handle;
    if !isempty (idx = find (oldax == match))
      ax = match(idx+1);

      strings = {};

      ## check every legend item candidates created after current axes
      ittyp = {"line", "surface"};
      idx += 1;
      hdls = match (idx:2:end);
      for ii = 1:length (hdls);
        hdl = hdls(ii);
        if any (strcmpi (get(hdl).type, {"line", "surface"}))
          str = get (hdl, "displayname");
          strings = [strings str];
        elseif strcmpi (get(hdl).type, "hggroup")
          if isprop (hdl, "bargroup")
            str = get (get (hdl).children(1), "displayname");
          elseif isprop (hdl, "format")
            ## FIXME - here the string is empty (the displayname is in
            ## children(2) but it still works!??
            str = get (get (hdl).children(1), "displayname");
          endif
          strings = [strings str];
        endif
      endfor

      posprop = get (ax, "activepositionproperty");
      location = hgS.properties.location;

      textpos = hgS.properties.textposition;

      show = "on";
      ## show = hgS.properties.visible;

      box = "boxoff";
      if strcmp (hgS.properties.box, "on");
        box = "boxon";
      endif
      h = legend (ax, strings, "location", location);
      ##h = legend (textpos);
      legend (box);
      if !strcmp (show, "on");
        legend ("hide");
      endif

      hgS.properties = rmfield (hgS.properties, {"textposition", "visible"});
      set (ax, "activepositionproperty", posprop);
      hgS.children= [];
    else
      error ("hdl2struct: didn't find an object")
    endif

  ## colorbar
  elseif strcmpi (hgS.properties.tag, "colorbar")
    oldax = hgS.properties.axes;
    if !isempty (idx = find (oldax == match))
      ax = match(idx+1);
      location = hgS.properties.location;
      h = colorbar ("peer", ax, location);
      hgS.children= [];
    else
      error ("hdl2struct: didn't find an object")
    endif

  endif
  hgSout = hgS;
endfunction ##createaxes

function [h] = createline (hgS);
  h = line ();
endfunction ##createline

function [h] = createpatch (hgS);
  h = patch ();
endfunction ##createpatch

function [h] = createtext (hgS);
  h = text ();
endfunction ##createtext

function [h] = createimage (hgS);
  h = image();
endfunction ##createimage

function [h] = createsurface (hgS);
  h = surface();
endfunction ##createsurface


function [h, hgSout, matchout] = createhg (hgS, matchin)
  ## Here we guess from properties the type of hggroup we should build
  ## an call corresponding high level functions
  ## We manually set "hold on" to avoid next hggroup be deleted
  ## the proper value of axes "nextplot" will finally be recovered

  hold on;

  fields = hgS.properties;

  if isfield (fields, "contourmatrix")
    ## contours
    xdata = hgS.properties.xdata;
    ydata = hgS.properties.ydata;
    zdata = hgS.properties.zdata;
    levellist = hgS.properties.levellist;
    textlist = hgS.properties.textlist;

    ## contour creation
    if strcmpi (hgS.properties.fill, "on")
      [cm2, h] = contourf (xdata, ydata, zdata, levellist);
    else
      [cm2, h] = contour (xdata, ydata, zdata, levellist);
    endif

    ## labels
    if strcmpi (hgS.properties.showtext, "on")
      clabel (cm2, h, textlist);
    endif

    ## delete already set properties and children
    hgS.properties = rmfield (hgS.properties, ...
                              {"xdata", "ydata", "zdata", ...
                               "contourmatrix", "levellist", ...
                               "fill", "labelspacing", ...
                               "levellistmode", "levelstep", ...
                               "levelstepmode", "textlist"...
                               "textlistmode" , "textstep", ...
                               "textstepmode", "zlevel", ...
                               "zlevelmode"});

  elseif isfield (fields, "udata") && isfield (fields, "vdata")
    ## quiver
    xdata = hgS.properties.xdata;
    ydata = hgS.properties.ydata;

    udata = hgS.properties.udata;
    vdata = hgS.properties.vdata;

    h = quiver (xdata, ydata, udata, vdata);

    ## delete already set properties and children
    hgS.properties = rmfield (hgS.properties, ...
                              {"xdata", "ydata", "zdata", ...
                               "xdatasource", "ydatasource", "zdatasource", ...
                               "udata", "vdata", "wdata", ...
                               "udatasource", "vdatasource", "wdatasource"});

  elseif isfield (fields, "format")
    ##errorbar
    form = hgS.properties.format;
    xdata = hgS.properties.xdata;
    ydata = hgS.properties.ydata;
    xldata = hgS.properties.xldata;
    ldata = hgS.properties.ldata;
    xudata = hgS.properties.xudata;
    udata = hgS.properties.udata;

    switch form
      case "xerr"
        h = errorbar (xdata, ydata, xldata, xudata, ">");
      case "yerr"
        h = errorbar (xdata, ydata, ldata, udata, "~");
      case "xyerr"
        h = errorbar (xdata, ydata, xldata, xudata, ldata, udata, "~>");
      case "box"
        h = errorbar (xdata, ydata, xldata, xudata, "#");
      case "boxy"
        h = errorbar (xdata, ydata, ldata, udata, "#~");
      case "boxxy"
        h = errorbar (xdata, ydata, xldata, xudata, ldata, udata, "#~>");
      otherwise
        error ("struct2hdl: couldn't guess the errorbar format")
    endswitch
    ## delete already set properties
    hgS.properties = rmfield (hgS.properties, ...
                              {"xdata", "ydata", ...
                               "xldata", "ldata", ...
                               "xudata", "udata", ...
                               "xldatasource", "ldatasource", ...
                               "xudatasource", "udatasource", ...
                               "format"});

  elseif isfield (fields, "bargroup")
    ## bar plot
    ## FIXME - here we don't have access to brothers so we first create all
    ## the barseries of the bargroup (but the last), then retrieve information,
    ## and rebuild the whole bargroup.
    ## The duplicate will be deleted in setprops when all info are retrieved

    bargroup = hgS.properties.bargroup;
    oldh = hgS.handle;

    temp = arrayfun (@(x) any (x == bargroup), [matchin(1:2:end) oldh]);
    tst = sum (temp) == length (bargroup);

    if (isscalar (bargroup) || !tst)
      xdata = hgS.properties.xdata;
      ydata = hgS.properties.ydata;

      h = bar (xdata, ydata);

      ## delete already set properties and children,
      hgS.properties = rmfield (hgS.properties, ...
                                {"xdata", "ydata", ...
                                 "xdatasource", "ydatasource", ...
                                 "bargroup",...
                                 "barwidth", "baseline"});

    else
      xdata = [];
      ydata = [];

      ##build x/y matrix
      nbar = length (bargroup);
      strings = {};
      for ii = 1:(nbar - 1)
        idx = find (matchin == bargroup(ii)) +1;
        hdl = matchin (idx);
        xdata = [xdata get(hdl).xdata];
        ydata = [ydata get(hdl).ydata];
      endfor

      xdata = [xdata hgS.properties.xdata];
      ydata = [ydata hgS.properties.ydata];
      width = hgS.properties.barwidth;
      h = bar (ydata, width);

      ## replace previous handles in match, add displayname
      matchin (2,((end-nbar+2):end)) = h (1:(end-1));
      h = h (end);
      ## delete already set properties and children,
      hgS.properties = rmfield (hgS.properties, ...
                                {"xdata", "ydata", "bargroup"...
                                 "barwidth", "baseline"});
    endif
  else
    error ("struct2hdl: couldn't guess the type of hggroup")
  endif

  hgSout = hgS;
  matchout = matchin;

endfunction

function setprops (hgS, h, matchin)
  if strcmpi (hgS.properties.tag, "")
    specs = hgS.children(hgS.special);
    hdls = arrayfun (@(x) x.handle, specs);
    nh = length(hdls);

    if !nh
      set (h, hgS.properties);
    else
      ## Specials are objects that where automatically constructed with
      ## current object. Among them are "x(yz)labels", "title", hggroup children


      fields = fieldnames(hgS.properties);
      vals = struct2cell (hgS.properties);
      idx = find (cellfun (@(x) valcomp(x, hdls) , vals));
      hgS.properties = rmfield (hgS.properties, fields(idx));

      set (h, hgS.properties);

      ## find  props with val == one of special handles
      nf = length (idx);
      fields = fields(idx);
      vals = vals(idx);
      while nf
        field = fields{nf};
        idx = find (hdls == vals{nf});
        spec = specs(idx);

        if isprop (h, field)
           h2 = get (h , field);
           set (h2, spec.properties)
        endif
        nf--;
      endwhile

      ## dirty fix for final positions of legend, colorbar and axes
      if strcmp (hgS.type, "axes")
        if !isempty (cb = findobj (get (h).parent,...
                                   "tag", "colorbar"));

          tst = arrayfun (@(x) any (h == get(x).axes), cb);
          if any (tst)
            cbnew = cb (find (tst));

            tst = find (cbnew == matchin) - 1;
            cbold = matchin (tst);

            tst = arrayfun (@(x) any (x.handle == cbold), hgS.children);
            idx = find (tst);

            checkset (cbnew, hgS.children(idx).properties)
          endif
        endif

        if !isempty (lg = findobj (get (h).parent,...
                                   "tag", "legend"));
          tst = arrayfun (@(x) any (h == get(x).userdata.handle), lg);
          if any (tst)
            lgnew = lg (find (tst));

            tst = find (lgnew == matchin) - 1;
            lgold = matchin (tst);

            tst = arrayfun (@(x) any (x.handle == lgold), hgS.children);
            idx = find (tst);

            checkset (lgnew, hgS.children(idx).properties)
          endif
        endif
        ## ensure that axes activepositionproperty is correct
        set (h, "activepositionproperty", ...
             hgS.properties.activepositionproperty);
      endif

      ## hggroup children are created by high level functions,
      ## copy their properties.  
      if strcmpi (hgS.type, "hggroup")
        if isfield (get(h), "bargroup")
          bg = get(h).bargroup;
          if numel (bg) > 1
            tmp = arrayfun (@(x) any (x == matchin (2:2:end)), bg);
            members = bg (find (tmp));
            ## copy previously constructed hggroups properties and
            ## delete redundant objects
            ncopy = numel (members) - 1;
            hg = findobj (gca, "type", "hggroup");
            for ii = 1:ncopy
              props = fieldnames (hgS.properties);
              ## copy props
              for kk = 1:numel (props)
                set (hg(ii+2), props{kk}, ...
                     get(hg(ii), props{kk}));
              endfor
              ## copy children
              chold = get(hg(ii)).children;
              chnew = get(hg(ii+ncopy)).children;
              nold = numel (chold);
              nnew = numel (chnew);
              if (nold == nnew)
                jj = 1;
                while jj <= nnew
                  try
                    set (chnew(jj), "displayname", ...
                         get(chold(jj)).displayname);
                  catch
                    sprintf ("struct2hdl: couldn't set hggroup children #%d props.", jj)
                  end_try_catch
                  jj++;
                endwhile
              else
                error ("struct2hdl: non-conformant number of children in hgggroup")
              endif
              delete (hg(ii));
            endfor
          endif
        endif

        nold = numel (hgS.children);
        nnew = numel (get(h).children);

        if (nold == nnew)
          hnew = get(h).children;
          ii = 1;
          while ii <= nnew
            try
              set (hnew (ii), "displayname", ...
                   hgS.children(ii).properties.displayname)
            catch
              sprintf ("struct2hdl: couldn't set hggroup children #%d props.", ii)
            end_try_catch
            ii ++;
          endwhile

        else
          error ("struct2hdl: non-conformant number of children in hgggroup")
        endif
      endif
    endif

  elseif strcmpi (hgS.properties.tag, "legend")
    oldax = hgS.properties.userdata.handle;
    if !isempty (idx = find (oldax == matchin))
      #remove all handles and properties such as "location" that redefines
      #the entire legend (and delete "h")
      hgS.properties = rmfield (hgS.properties, ...
                                {"userdata", "xlabel",...
                                 "ylabel", "zlabel", "location", ...
                                 "title", "string", "tag", "box"});
     set (h, hgS.properties);
    else
      error ("hdl2struct: didn't find an object")
    endif

  elseif strcmpi (hgS.properties.tag, "colorbar")
    oldax = hgS.properties.axes;
    if !isempty (idx = find (oldax == matchin))
      hgS.properties = rmfield (hgS.properties, ...
                                {"userdata", "xlabel" ...
                                 "ylabel", "zlabel", "location", ...
                                 "title", "axes", "tag"});
      set (h, hgS.properties);
    else
      error ("hdl2struct: didn't find an object")
    endif

  endif

endfunction ##setprops

function out = valcomp (x, hdls)

  if isfloat(x) && isscalar(x)
    out = any (x == hdls);
  else
    out = 0;
  endif
endfunction ## valcomp

function [out] = gencmp (val1, val2)
  if  (numel (val1) != numel (val2))
    out = false;
  elseif (ishandle (val1) || ishandle (val2))
    ## omit handles
    out = true;
  elseif (isnumeric (val1) && isnumeric (val2) )
    out = all (val1 == val2);
  elseif (ischar (val1) && ischar (val2)
          || iscellstr (val1) && iscellstr (val2))
    out = all (strcmpi (val1, val2));
  else
    out = false;
  endif
endfunction ## gencmp

function checkset (h1, propval)

  s1 = get (h1);
  s2 = propval;

  tmp1 = sort (fieldnames (s1));
  tmp2 = sort (fieldnames (s2));
  union1 = cellfun (@(x) any (strcmp (x, tmp2)), tmp1);
  union2 = cellfun (@(x) any (strcmp (x, tmp1)), tmp2);

  ## nfields
  strout.nfields = [numel(tmp1), numel(tmp2)];

  ## union
  strout.union = tmp1 (find (union1));

  ## exclude2
  strout.exclude2 = tmp1 (find (union1 == 0));

  ## exclude1
  strout.exclude1 = tmp2 (find (union2 == 0));

  ## valdiff
  forget = {"deletefcn", "userdata"};
  strout.valdiff = {};
  nu = numel (strout.union);
  for ii = 1:nu
    field = strout.union{ii};
    if (! gencmp (s1.(field), s2.(field))
        && ! any (strcmp (field, forget)))
      strout.valdiff{1}.(field) = s1.(field);
      strout.valdiff{2}.(field) = s2.(field);
    endif
  endfor

  if !isempty (strout.valdiff)
    set (h1, strout.valdiff{2})
  endif

endfunction ## checkset
